@@ -515,7 +515,9 @@ void get_tboot_extpol(void)
     }
 
     if ( tb_strcmp(extpol, "agile") == 0 ) {
-        tpm->extpol = TB_EXTPOL_AGILE;
+        printk(TBOOT_WARN"Warning: \"agile\" selected in extpol, "
+                         "this option is deprecated\n");
+        tpm->extpol = TB_EXTPOL_FIXED;
         tpm->cur_alg = TB_HALG_SHA256;
     } else if ( tb_strcmp(extpol, "embedded") == 0 ) {
         tpm->extpol = TB_EXTPOL_EMBEDDED;
