@@ -70,6 +70,8 @@
 #include <integrity.h>
 #include <cmdline.h>
 #include <tpm_20.h>
+#include <vtd.h>
+#include <efi_memmap.h>
 
 extern void _prot_to_real(uint32_t dist_addr);
 extern bool set_policy(void);
@@ -137,21 +139,36 @@ static void copy_s3_wakeup_entry(void)
     }
 
     /* backup target address space first */
-    memcpy(g_saved_s3_wakeup_page, (void *)TBOOT_S3_WAKEUP_ADDR,
+    tb_memcpy(g_saved_s3_wakeup_page, (void *)TBOOT_S3_WAKEUP_ADDR,
            s3_wakeup_end - s3_wakeup_16);
 
     /* copy s3 entry into target mem */
-    memcpy((void *)TBOOT_S3_WAKEUP_ADDR, s3_wakeup_16,
+    tb_memcpy((void *)TBOOT_S3_WAKEUP_ADDR, s3_wakeup_16,
            s3_wakeup_end - s3_wakeup_16);
 }
 
 static void restore_saved_s3_wakeup_page(void)
 {
     /* restore saved page */
-    memcpy((void *)TBOOT_S3_WAKEUP_ADDR, g_saved_s3_wakeup_page,
+    tb_memcpy((void *)TBOOT_S3_WAKEUP_ADDR, g_saved_s3_wakeup_page,
            s3_wakeup_end - s3_wakeup_16);
 }
 
+static inline void print_tboot_shared(const tboot_shared_t *tboot_shared)
+{
+    printk(TBOOT_DETA"tboot_shared data:\n");
+    printk(TBOOT_DETA"\t version: %d\n", tboot_shared->version);
+    printk(TBOOT_DETA"\t log_addr: 0x%08x\n", tboot_shared->log_addr);
+    printk(TBOOT_DETA"\t shutdown_entry: 0x%08x\n", tboot_shared->shutdown_entry);
+    printk(TBOOT_DETA"\t shutdown_type: %d\n", tboot_shared->shutdown_type);
+    printk(TBOOT_DETA"\t tboot_base: 0x%08x\n", tboot_shared->tboot_base);
+    printk(TBOOT_DETA"\t tboot_size: 0x%x\n", tboot_shared->tboot_size);
+    printk(TBOOT_DETA"\t num_in_wfs: %u\n", tboot_shared->num_in_wfs);
+    printk(TBOOT_DETA"\t flags: 0x%8.8x\n", tboot_shared->flags);
+    printk(TBOOT_DETA"\t ap_wake_addr: 0x%08x\n", (uint32_t)tboot_shared->ap_wake_addr);
+    printk(TBOOT_DETA"\t ap_wake_trigger: %u\n", tboot_shared->ap_wake_trigger);
+}
+
 static void post_launch(void)
 {
     uint64_t base, size;
@@ -169,7 +186,8 @@ static void post_launch(void)
     txt_post_launch();
 
     /* backup DMAR table */
-    save_vtd_dmar_table();
+    if ( get_tboot_save_vtd() )
+        vtd_save_dmar_table();
 
     if ( s3_flag  )    
          s3_launch();
@@ -206,6 +224,9 @@ static void post_launch(void)
     printk(TBOOT_INFO"protecting tboot (%Lx - %Lx) in e820 table\n", base,      (base + size - 1));
     if ( !e820_protect_region(base, size, mem_type) )      
         apply_policy(TB_ERR_FATAL);
+    if (!efi_memmap_reserve(base, size)) {
+        apply_policy(TB_ERR_FATAL);
+    }
 
     /*
      * verify modules against policy
@@ -232,7 +253,7 @@ static void post_launch(void)
 	/*
      * init MLE/kernel shared data page
      */
-    memset(&_tboot_shared, 0, PAGE_SIZE);
+    tb_memset(&_tboot_shared, 0, PAGE_SIZE);
     _tboot_shared.uuid = (uuid_t)TBOOT_SHARED_UUID;
     _tboot_shared.version = 6;
     _tboot_shared.log_addr = (uint32_t)g_log;
@@ -335,20 +356,29 @@ void begin_launch(void *addr, uint32_t magic)
            // cmdline = skip_filename(cmdline_orig);
             cmdline = cmdline_orig;
         }
-        memset(g_cmdline, '\0', sizeof(g_cmdline));
+        tb_memset(g_cmdline, '\0', sizeof(g_cmdline));
         if (cmdline)
-            strncpy(g_cmdline, cmdline, sizeof(g_cmdline)-1);
+            tb_strncpy(g_cmdline, cmdline, sizeof(g_cmdline)-1);
     }
 
     /* always parse cmdline */
     tboot_parse_cmdline();
 
     /* initialize all logging targets */
-    printk_init();
+    if ((is_launched() && get_framebuffer_info(g_ldr_ctx) != NULL) || s3_flag) {
+        /* 
+         * We are using framebuffer method for printing VGA logs. It does not
+         * when PMRs are programmed, so in post-launch we have to disable VGA
+         * logging.
+         */
+        printk_init(true);
+    } else {
+        printk_init(false);
+    }
 
-    printk(TBOOT_INFO"******************* TBOOT *******************\n");
+    printk(TBOOT_INFO"*********************** TBOOT ***********************\n");
     printk(TBOOT_INFO"   %s\n", TBOOT_CHANGESET);
-    printk(TBOOT_INFO"*********************************************\n");
+    printk(TBOOT_INFO"*****************************************************\n");
 
     printk(TBOOT_INFO"command line: %s\n", g_cmdline);
     /* if telled to check revocation acm result, go with simplified path */
@@ -376,6 +406,12 @@ void begin_launch(void *addr, uint32_t magic)
     /* make copy of e820 map that we will use and adjust */
     if ( !s3_flag ) {
         if ( !copy_e820_map(g_ldr_ctx) )  apply_policy(TB_ERR_FATAL);
+        if (efi_memmap_copy(g_ldr_ctx)) {
+            if (get_tboot_dump_memmap()) {
+                printk(TBOOT_INFO"Original EFI memory map:\n");
+                efi_memmap_dump();
+            }
+        }
     }
 
     /* we need to make sure this is a (TXT-) capable platform before using */
@@ -429,7 +465,7 @@ void begin_launch(void *addr, uint32_t magic)
     /* this is being called post-measured launch */
     if ( is_launched() ){
         printk(TBOOT_INFO"Post_launch started ...\n");
-	post_launch();
+        post_launch();
     }
 
     /* make the CPU ready for measured launch */
@@ -473,7 +509,8 @@ void s3_launch(void)
     }
 
     /* remove DMAR table if necessary */
-    remove_vtd_dmar_table();
+    if ( get_tboot_save_vtd() )
+        vtd_remove_dmar_table();
 
     if ( !is_launched() )
         apply_policy(TB_ERR_S3_INTEGRITY);
@@ -485,6 +522,15 @@ void s3_launch(void)
     }
 
     print_tboot_shared(&_tboot_shared);
+    
+    if (g_tpm_family != TPM_IF_20_CRB ) {
+        if (!release_locality(tpm->cur_loc))
+            printk(TBOOT_ERR"Release TPM FIFO locality %d failed \n", tpm->cur_loc);
+    }
+    else {
+        if (!tpm_relinquish_locality_crb(tpm->cur_loc))
+            printk(TBOOT_ERR"Relinquish TPM CRB locality %d failed \n", tpm->cur_loc);
+    }
 
     /* (optionally) pause when transferring kernel resume */
     if ( g_vga_delay > 0 )
@@ -501,9 +547,9 @@ static void shutdown_system(uint32_t shutdown_type)
     char type[32];
 
     if ( shutdown_type >= ARRAY_SIZE(types) )
-        snprintf(type, sizeof(type), "unknown: %u", shutdown_type);
+        tb_snprintf(type, sizeof(type), "unknown: %u", shutdown_type);
     else {
-        strncpy(type, types[shutdown_type], sizeof(type));
+        tb_strncpy(type, types[shutdown_type], sizeof(type));
         type[sizeof(type) - 1] = '\0';
     }
     printk(TBOOT_INFO"shutdown_system() called for shutdown_type: %s\n", type);
@@ -591,7 +637,8 @@ void shutdown(void)
 
     if ( _tboot_shared.shutdown_type == TB_SHUTDOWN_S3 ) {
         /* restore DMAR table if needed */
-        restore_vtd_dmar_table();
+        if ( get_tboot_save_vtd() )
+            vtd_restore_dmar_table();
 	if ( tpm->major == TPM20_VER_MAJOR ) {
 	    tpm_fp->context_flush(tpm, tpm->cur_loc, handle2048);
 	    tpm_fp->context_load(tpm, tpm->cur_loc, &tpm2_context_saved, &handle2048);
@@ -609,20 +656,26 @@ void shutdown(void)
                error */
         else
             /* wipe S3 key from memory now that it is sealed */
-            memset(_tboot_shared.s3_key, 0, sizeof(_tboot_shared.s3_key));
+            tb_memset(_tboot_shared.s3_key, 0, sizeof(_tboot_shared.s3_key));
     }
 
     /* cap dynamic PCRs extended as part of launch (17, 18, ...) */
     if ( is_launched() ) {
 
         /* cap PCRs to ensure no follow-on code can access sealed data */
-        tpm_fp->cap_pcrs(tpm, tpm->cur_loc, -1);
+        if (!tpm_fp->cap_pcrs(tpm, tpm->cur_loc, -1)) {
+            printk(TBOOT_ERR"failed to cap PCRs\n");
+            apply_policy(TB_ERR_FATAL);
+        }
 
         /* have TPM save static PCRs (in case VMM/kernel didn't) */
         /* per TCG spec, TPM can invalidate saved state if any other TPM
            operation is performed afterwards--so do this last */
-        if ( _tboot_shared.shutdown_type == TB_SHUTDOWN_S3 )
-            tpm_fp->save_state(tpm, tpm->cur_loc);
+        if ( _tboot_shared.shutdown_type == TB_SHUTDOWN_S3 ) {
+            if (tpm_fp->save_state(tpm, tpm->cur_loc) != 0) {
+                printk(TBOOT_ERR"failed to save TPM state\n");
+            }
+        }
 
         /* scrub any secrets by clearing their memory, then flush cache */
         /* we don't have any secrets to scrub, however */
